@@ -11,6 +11,9 @@
 
 /*------------------------------------------------------------------------*/
 
+/* NB:  int nr_limbs = R_mpfr_nr_limbs(r)  [ in MPFR_as_R() ]  or
+                     = N_LIMBS(i_prec)     [ in d2mpfr1_()  ]
+*/
 #if GMP_NUMB_BITS == 32
 # define R_mpfr_nr_ints nr_limbs
 # define R_mpfr_exp_size 1
@@ -21,12 +24,13 @@
 # error "R <-> C Interface *not* implemented for GMP_NUMB_BITS=" ## GMP_NUMB_BITS
 #endif
 
-#define R_mpfr_MPFR_2R_init(_V_)					\
+// Initialize contents (4 slots) of a "mpfr1" R object
+#define R_mpfr_MPFR_2R_init(_V_, _d_length_)				\
     SEXP _V_ = PROTECT(NEW_OBJECT(MAKE_CLASS("mpfr1")));		\
-    SEXP prec_R = ALLOC_SLOT(_V_, Rmpfr_precSym, INTSXP, 1);		\
-    SEXP sign_R = ALLOC_SLOT(_V_, Rmpfr_signSym, INTSXP, 1);		\
-    SEXP exp_R  = ALLOC_SLOT(_V_, Rmpfr_expSym,  INTSXP, R_mpfr_exp_size); \
-    SEXP d_R    = ALLOC_SLOT(_V_, Rmpfr_d_Sym,   INTSXP, R_mpfr_nr_ints); \
+    SEXP prec_R = PROTECT(ALLOC_SLOT(_V_, Rmpfr_precSym, INTSXP, 1));	\
+    SEXP sign_R = PROTECT(ALLOC_SLOT(_V_, Rmpfr_signSym, INTSXP, 1));	\
+    SEXP exp_R  = PROTECT(ALLOC_SLOT(_V_, Rmpfr_expSym,  INTSXP, R_mpfr_exp_size)); \
+    SEXP d_R    = PROTECT(ALLOC_SLOT(_V_, Rmpfr_d_Sym,   INTSXP, _d_length_)); \
     /* the integer vector which makes up the mantissa: */		\
     int *dd = INTEGER(d_R),						\
 	*ex = INTEGER(exp_R) /* the one for the exponent */
@@ -34,6 +38,8 @@
 /*------------------------*/
 #if GMP_NUMB_BITS == 32
 /*                  ---- easy : a gmp-limb is an int <--> R */
+
+// This is only ok  if( mpfr_regular_p(.) ), i.e. not for {0, NaN, Inf}:
 # define R_mpfr_FILL_DVEC(i)					\
     R_mpfr_dbg_printf(2,"r..d[i=%d] = 0x%lx\n",i,r->_mpfr_d[i]); \
     dd[i] = (int) r->_mpfr_d[i]
@@ -57,6 +63,7 @@
 //                                                   1  4   8|  4   8
 # define LEFT_SHIFT(_LONG_) (((unsigned long long)(_LONG_)) << 32)
 
+// This is only ok  if( mpfr_regular_p(.) ), i.e. not for {0, NaN, Inf}:
 # define R_mpfr_FILL_DVEC(i)						\
     R_mpfr_dbg_printf(2,"r..d[i=%d] = 0x%lx\n",i,r->_mpfr_d[i]);	\
     dd[2*i]  = (int) RIGHT_HALF(r->_mpfr_d[i]);				\
@@ -84,49 +91,52 @@
 
 
 
-
 #define R_mpfr_MPFR_2R_fill			\
     /* now fill the slots of val */		\
     INTEGER(prec_R)[0] = (int)r->_mpfr_prec;	\
     INTEGER(sign_R)[0] = (int)r->_mpfr_sign;	\
     R_mpfr_FILL_EXP;				\
-    /* the full *vector* of limbs : */		\
-    for(i=0; i < nr_limbs; i++) {		\
-	R_mpfr_FILL_DVEC(i);			\
+    if(regular_p) {				\
+	/* the full *vector* of limbs : */	\
+	for(i=0; i < nr_limbs; i++) {		\
+            R_mpfr_FILL_DVEC(i);		\
+	}					\
     }
 
 /* Return an R "mpfr1" object corresponding to mpfr input: */
 SEXP MPFR_as_R(mpfr_t r) {
 
-    int nr_limbs = R_mpfr_nr_limbs(r), i;
+    int nr_limbs = R_mpfr_nr_limbs(r),
+	regular_p = mpfr_regular_p(r), i;
 
-    R_mpfr_MPFR_2R_init(val);
+    R_mpfr_MPFR_2R_init(val, (regular_p ? R_mpfr_nr_ints : 0));
 
     R_mpfr_MPFR_2R_fill;
 
-    UNPROTECT(1);
+    UNPROTECT(5);
     return val;
 }
 
 SEXP d2mpfr1_(double x, int i_prec, mpfr_rnd_t rnd)
 {
     mpfr_t r;
-    int nr_limbs = N_LIMBS(i_prec), i;
+    int nr_limbs = N_LIMBS(i_prec), regular_p, i;
 
     R_mpfr_check_prec(i_prec);
 
-    R_mpfr_MPFR_2R_init(val);
-
     mpfr_init2 (r, (mpfr_prec_t)i_prec);
     mpfr_set_d (r, x, rnd);
 
+    regular_p = mpfr_regular_p(r);
+    R_mpfr_MPFR_2R_init(val, (regular_p ? R_mpfr_nr_ints : 0));
+
     R_mpfr_MPFR_2R_fill;
 
     /* free space used by the MPFR variables */
     mpfr_clear (r);
     mpfr_free_cache(); /* <- Manual 4.8 "Memory Handling" strongly advises ...*/
 
-    UNPROTECT(1);
+    UNPROTECT(5);
     return val;
 }/* d2mpfr1_ */
 
@@ -289,10 +299,11 @@ void R_asMPFR(SEXP x, mpfr_ptr r)
 
     int x_prec = INTEGER(prec_R)[0],
 	nr_limbs = N_LIMBS(x_prec), i;
+    Rboolean regular_x = length(d_R) > 0;
     int *dd = INTEGER(d_R),/* the vector which makes up the mantissa */
 	*ex = INTEGER(exp_R), ex1; /* the one for the exponent */
 
-    if(length(d_R) != R_mpfr_nr_ints)
+    if(regular_x && length(d_R) != R_mpfr_nr_ints)
 	error("nr_limbs(x_prec)= nr_limbs(%d)= %d : length(<d>) == %d != R_mpfr_nr_ints == %d",
 	      x_prec, nr_limbs, length(d_R), R_mpfr_nr_ints);
     if(length(exp_R) < R_mpfr_exp_size) {
@@ -305,10 +316,11 @@ void R_asMPFR(SEXP x, mpfr_ptr r)
     mpfr_set_prec(r, (mpfr_prec_t) x_prec);
     r->_mpfr_sign = (mpfr_sign_t) INTEGER(GET_SLOT(x, Rmpfr_signSym))[0];
     R_mpfr_GET_EXP;
-    /* the full *vector* of limbs : */
-    for(i=0; i < nr_limbs; i++) {
-	R_mpfr_GET_DVEC(i);
-    }
+    if(regular_x)
+	/* the full *vector* of limbs : */
+	for(i=0; i < nr_limbs; i++) {
+	    R_mpfr_GET_DVEC(i);
+	}
     return;
 }
 
@@ -417,38 +429,59 @@ SEXP mpfr2i(SEXP x, SEXP rnd_mode) {
 }
 
 /* Convert R "mpfr" object (list of "mpfr1")  to R "character" vector,
- * using precision 'prec' which can be NA/NULL in which case
- * "full precision" (as long as necessary) is used : */
-SEXP mpfr2str(SEXP x, SEXP digits, SEXP base) {
+ * using 'digits' (or determinining it):
+ *  1) digits = NULL , maybe_full = FALSE (<==> 'scientific = TRUE')
+ *     --------------- -------------~~~~~ ==> set digits  <=>  getPrec(x)
+
+ *  2) digits = NULL , maybe_full = TRUE (<=> 'scientific' = NA or FALSE  in the calling formatMpfr()
+ *     -------------   ----------------- ==> set digits  <=>  max(getPrec(x), #{"digits left of '.'"})
+
+ *  3) digits = <num>, maybe_full = TRUE (<=> 'scientific' = TRUE  in the calling formatMpfr()
+ *     -------------   -----------------==> set digits  <=>  max(digit, getPrec(x), #{"digits left of '.'"}))
+
+ */
+SEXP mpfr2str(SEXP x, SEXP digits, SEXP maybeFull, SEXP base) {
     int n = length(x), i;
-    int n_dig = isNull(digits) ? 0 : asInteger(digits);
-    int dig_n_max = -1;
-    SEXP val = PROTECT(allocVector(VECSXP, 4)),
-	nms, str, exp, fini, zero;
-    int *i_exp, *is_fin, *is_0;
     int B = asInteger(base); // = base for output
-    double p_fact = (B == 2) ? 1. : log(B) / M_LN2;
+    int n_dig = isNull(digits) ? 0 : asInteger(digits);
+    if(n_dig < 0) error("'digits' must be NULL or a positive integer");
+    Rboolean maybe_full = asLogical(maybeFull);
+    if(maybe_full == NA_LOGICAL) // cannot happen when called "regularly"
+	error("'maybe.full' must be TRUE or FALSE");
+
+    R_mpfr_dbg_printf(1,"mpfr2str(*, digits=%d, maybeF=%s, base=%d): ",
+		      n_dig, (maybe_full ? "TRUE" : "False"), B);
+
+    /* int dig_n_max = -1; */
+    /* SEXP val = PROTECT(allocVector(VECSXP, 4)), */
+    /* 	nms, str, exp, fini, zero; */
+    /* int *i_exp, *is_fin, *is_0; */
     char *ch = NULL;
-    mpfr_t R_i;
 
-    if(n_dig < 0)
-	error("'digits' must be NULL or integer >= 0");
+    /* N_digits == 1 , for base = 2, 4, 8, 16, 32 (base <= 62 !) gives bad abort from MPFR:
+       get_str.c:2306: MPFR assertion failed: m >= 2 || ((((b) & ((b) - 1)) == 0) == 0 && m >= 1)
+       ...  Aborted ... (Speicherabzug geschrieben)
 
-    /* be "overprotective" for now ... */
-    SET_VECTOR_ELT(val, 0, str = PROTECT(allocVector(STRSXP, n)));
-    SET_VECTOR_ELT(val, 1, exp = PROTECT(allocVector(INTSXP, n)));
-    SET_VECTOR_ELT(val, 2, fini= PROTECT(allocVector(LGLSXP, n)));
-    SET_VECTOR_ELT(val, 3, zero= PROTECT(allocVector(LGLSXP, n)));
-    nms = PROTECT(allocVector(STRSXP, 4));
-    SET_STRING_ELT(nms, 0, mkChar("str"));
-    SET_STRING_ELT(nms, 1, mkChar("exp"));
-    SET_STRING_ELT(nms, 2, mkChar("finite"));
-    SET_STRING_ELT(nms, 3, mkChar("is.0"));
+      the MPFR doc (see mpfr_get_str below) *says* N >= 2 is required,
+      but we have used N = 1 for B = 10 a lot in the past ! */
+
+    Rboolean base_is_2_power = (B == 2 || B == 4 || B == 8 || B == 16 || B == 32);
+    Rboolean n_dig_1_problem = (n_dig == 1) && base_is_2_power;
+    size_t N_digits = n_dig_1_problem ? 2 : n_dig;
+    SEXP
+	val = PROTECT(allocVector(VECSXP, 4)),
+	nms = PROTECT(allocVector(STRSXP, 4)), str, exp, fini, zero;
+    SET_VECTOR_ELT(val, 0, str = PROTECT(allocVector(STRSXP, n))); SET_STRING_ELT(nms, 0, mkChar("str"));
+    SET_VECTOR_ELT(val, 1, exp = PROTECT(allocVector(INTSXP, n))); SET_STRING_ELT(nms, 1, mkChar("exp"));
+    SET_VECTOR_ELT(val, 2, fini= PROTECT(allocVector(LGLSXP, n))); SET_STRING_ELT(nms, 2, mkChar("finite"));
+    SET_VECTOR_ELT(val, 3, zero= PROTECT(allocVector(LGLSXP, n))); SET_STRING_ELT(nms, 3, mkChar("is.0"));
     setAttrib(val, R_NamesSymbol, nms);
-    i_exp = INTEGER(exp);
-    is_fin= LOGICAL(fini);
-    is_0  = LOGICAL(zero);
-
+    int *i_exp = INTEGER(exp),
+	*is_fin= LOGICAL(fini),
+	*is_0  = LOGICAL(zero);
+    double p_fact = (B == 2) ? 1. : log(B) / M_LN2;
+    int dig_n_max = -1; // := max_i { dig_needed[i] }
+    mpfr_t R_i;
     mpfr_init(R_i); /* with default precision */
 
     for(i=0; i < n; i++) {
@@ -463,27 +496,56 @@ SEXP mpfr2str(SEXP x, SEXP digits, SEXP base) {
  * Originally hoped it was solvable via  R_alloc() etc, but it seems the problem is
  * deeper and I currently suspect a problem/bug in MPFR library's  mpfr_get_str(..) */
 	ch = mpfr_get_str(NULL, exp_ptr, B,
-			  (size_t) n_dig, R_i, MPFR_RNDN);
+			  (size_t) N_digits, R_i, MPFR_RNDN);
 #else
 	if(n_dig) {/* use it as desired precision */
-	    dig_needed = n_dig;
-	} else { /* n_dig = 0 --> string will use "enough" digits */
-	    dig_needed = p_fact * (int)R_i->_mpfr_prec;
+	    dig_needed = N_digits;
+	    R_mpfr_dbg_printf(1," [i=%d]: ... -> dig.n = %d ", i, dig_needed);
+	} else { /* N_digits = 0 --> string must use "enough" digits */
+	    double need_dig =
+		ceil(fmax2((double)R_i->_mpfr_prec,
+			   // when prec is too small:
+			   (double)mpfr_get_exp(R_i)) / p_fact);
+	    if(need_dig > 268435456 /* = 2^28 */) // << FIXME, somewhat arbitrary
+		error(_(".mpfr2str(): too large 'need_dig'; please set 'digits = <number>'"));
+// FIXME: rather set   maybe_full = FALSE  (???)
+	    dig_needed = (int) need_dig;
+	    R_mpfr_dbg_printf(1," [i=%d]: prec=%ld, exp2=%ld -> (n.dig,dig.n)=(%g,%d) ",
+			      i, R_i->_mpfr_prec, mpfr_get_exp(R_i),
+			      need_dig, dig_needed);
+	    if(dig_needed <= 1 && base_is_2_power) { // have n_dig_problem:
+		R_mpfr_dbg_printf(1," [i=%d]: base_is_2_power & dig_needed=%d ==> fudge dig_n. := 2",
+				  i, dig_needed);
+		dig_needed = 2;
+	    }
 	}
 	if (i == 0) { /* first time */
 	    dig_n_max = dig_needed;
-	    ch = (char *) R_alloc(dig_needed + 2, sizeof(char));
+	    ch = (char *) R_alloc(imax2(dig_n_max + 2, 7), // 7 : '-@Inf@' (+ \0)n_str,
+				  sizeof(char));
 	}
-	else if(!n_dig && dig_needed > dig_n_max) {
-	    ch = (char *) S_realloc(ch, dig_needed + 2, dig_n_max + 2,
+	else if(!N_digits && dig_needed > dig_n_max) { // enlarge :
+	    ch = (char *) S_realloc(ch,
+				    imax2(dig_needed + 2, 7),
+				    imax2(dig_n_max  + 2, 7),
 				    sizeof(char));
 	    dig_n_max = dig_needed;
 	}
 
-	/* char * mpfr_get_str (char *STR, mpfr_exp_t *EXPPTR, int B,
-	 *			size_t N, mpfr_t OP, mpfr_rnd_t RND) */
+	/*  char* mpfr_get_str (char *STR, mpfr_exp_t *EXPPTR, int B,
+	 *			size_t N, mpfr_t OP, mpfr_rnd_t RND)
+
+	 Convert OP to a string of digits in base B, with rounding in the
+	 direction RND, where N is either zero (see below) or the number of
+	 significant digits output in the string; in the latter case, N must
+	 be greater or equal to 2.  The base may vary from 2 to 62;
+	 .........
+	 .........  ==> MPFR info manual  "5.4 Conversion Functions"
+	*/
+	R_mpfr_dbg_printf(1," .. dig_n_max=%d\n", dig_n_max);
+
 	mpfr_get_str(ch, exp_ptr, B,
-		     (size_t) n_dig, R_i, MPFR_RNDN);
+		     (size_t) dig_n_max, R_i, MPFR_RNDN);
 #endif
 	SET_STRING_ELT(str, i, mkChar(ch));
 	i_exp[i] = (int) exp_ptr[0];
